@@ -100,7 +100,7 @@ typedef struct
     globus_xio_system_socket_t          fd;
     char *                              bind_address;
     globus_bool_t                       restrict_port;
-    globus_bool_t                       resuseaddr;
+    globus_bool_t                       reuseaddr;
     globus_bool_t                       no_ipv6;
     
     /* handle attrs */
@@ -159,7 +159,6 @@ static globus_mutex_t                   globus_l_xio_tcp_port_range_state_lock;
 /* string parse options table.  this rable has all of the options that
     the use can set via strings */
 static globus_xio_string_cntl_table_t tcp_l_string_opts_table[] =
-
 {
     {"port", GLOBUS_XIO_TCP_SET_SERVICE,
         globus_xio_string_cntl_string},
@@ -181,6 +180,21 @@ static globus_xio_string_cntl_table_t tcp_l_string_opts_table[] =
         globus_xio_string_cntl_bool},
     {NULL, 0, NULL}
 };
+
+static globus_xio_string_cntl_table_t tcp_l_handle_string_opts_table[] =
+{
+    {"keepalive", GLOBUS_XIO_TCP_SET_KEEPALIVE,
+        globus_xio_string_cntl_bool},
+    {"oobinline", GLOBUS_XIO_TCP_SET_OOBINLINE,
+        globus_xio_string_cntl_bool},
+    {"sndbuf", GLOBUS_XIO_TCP_SET_SNDBUF,
+        globus_xio_string_cntl_formated_int},
+    {"rcvbuf", GLOBUS_XIO_TCP_SET_RCVBUF,
+        globus_xio_string_cntl_formated_int},
+    {"nodelay", GLOBUS_XIO_TCP_SET_NODELAY,
+        globus_xio_string_cntl_bool},
+    {NULL, 0, NULL}
+};
 /*
  *  server structure
  */
@@ -277,7 +291,7 @@ globus_l_xio_tcp_file_lock(void)
 
 static
 void
-globus_l_xio_tcp_file_unlock()
+globus_l_xio_tcp_file_unlock(void)
 {
     int                                 rc;
     struct flock                        fl;
@@ -615,15 +629,15 @@ globus_l_xio_tcp_attr_cntl(
         *out_bool = attr->restrict_port;
         break;
       
-      /* globus_bool_t                  resuseaddr */
+      /* globus_bool_t                  reuseaddr */
       case GLOBUS_XIO_TCP_SET_REUSEADDR:
-        attr->resuseaddr = va_arg(ap, globus_bool_t);
+        attr->reuseaddr = va_arg(ap, globus_bool_t);
         break;
         
-      /* globus_bool_t *                resuseaddr_out */
+      /* globus_bool_t *                reuseaddr_out */
       case GLOBUS_XIO_TCP_GET_REUSEADDR:
         out_bool = va_arg(ap, globus_bool_t *);
-        *out_bool = attr->resuseaddr;
+        *out_bool = attr->reuseaddr;
         break;
       
       /* globus_bool_t                  no_ipv6 */
@@ -752,6 +766,123 @@ globus_l_xio_tcp_attr_cntl(
         *out_bool = attr->use_blocking_io;
         break;
 
+      case GLOBUS_XIO_GET_STRING_OPTIONS:
+      {
+        out_string = va_arg(ap, char **);
+        size_t string_opts_len = 1;
+        if (attr->listener_serv)
+        {
+            string_opts_len += snprintf(
+                NULL, 0, "port=%s;", attr->listener_serv);
+        }
+        if (attr->listener_min_port || attr->listener_max_port)
+        {
+            string_opts_len += snprintf(
+                NULL, 0,
+                "listen_range=%d %d;",
+                attr->listener_min_port,
+                attr->listener_max_port);
+        }
+       
+        if (attr->bind_address)
+        {
+            string_opts_len += snprintf(
+                NULL, 0,
+                "iface=%s;",
+                attr->bind_address);
+        }
+
+        string_opts_len += snprintf(
+                NULL, 0,
+                "reuse=%s;",
+                attr->reuseaddr ? "true" : "false");
+        string_opts_len += snprintf(
+                NULL, 0,
+                "noipv6=%s;",
+                attr->no_ipv6 ? "true" : "false");
+        string_opts_len += snprintf(
+                NULL, 0,
+                "keepalive=%s;",
+                attr->keepalive ? "true" : "false");
+        if (attr->sndbuf)
+        {
+            string_opts_len += snprintf(
+                    NULL, 0,
+                    "sndbuf=%d;",
+                    attr->sndbuf);
+        }
+        if (attr->rcvbuf)
+        {
+            string_opts_len += snprintf(
+                    NULL, 0,
+                    "rcvbuf=%d;",
+                    attr->rcvbuf);
+        }
+	string_opts_len += snprintf(
+                    NULL, 0,
+                    "nodelay=%s;",
+                    attr->nodelay ? "true" : "false");
+
+        *out_string = malloc(string_opts_len);
+
+        string_opts_len = 0;
+        if (attr->listener_serv)
+        {
+            string_opts_len += sprintf(
+                *out_string + string_opts_len,
+                "port=%s;", attr->listener_serv);
+        }
+        if (attr->listener_min_port || attr->listener_max_port)
+        {
+            string_opts_len += sprintf(
+                *out_string + string_opts_len,
+                "listen_range=%d %d;",
+                attr->listener_min_port,
+                attr->listener_max_port);
+        }
+       
+        if (attr->bind_address)
+        {
+            string_opts_len += sprintf(
+                *out_string + string_opts_len,
+                "iface=%s;",
+                attr->bind_address);
+        }
+
+        string_opts_len += sprintf(
+                *out_string + string_opts_len,
+                "reuse=%s;",
+                attr->reuseaddr ? "true" : "false");
+        string_opts_len += sprintf(
+                *out_string + string_opts_len,
+                "noipv6=%s;",
+                attr->no_ipv6 ? "true" : "false");
+        string_opts_len += sprintf(
+                *out_string + string_opts_len,
+                "keepalive=%s;",
+                attr->keepalive ? "true" : "false");
+        if (attr->sndbuf)
+        {
+            string_opts_len += sprintf(
+                    *out_string + string_opts_len,
+                    "sndbuf=%d;",
+                    attr->sndbuf);
+        }
+        if (attr->rcvbuf)
+        {
+            string_opts_len += sprintf(
+                    *out_string + string_opts_len,
+                    "rcvbuf=%d;",
+                    attr->rcvbuf);
+        }
+        string_opts_len += sprintf(
+                *out_string + string_opts_len,
+                "nodelay=%s;",
+                attr->nodelay ? "true" : "false");
+        *((*out_string) + string_opts_len - 1) = '\0';
+      }
+        break;
+
       default:
         result = GlobusXIOErrorInvalidCommand(cmd);
         goto error_invalid;
@@ -874,7 +1005,7 @@ globus_l_xio_tcp_apply_handle_attrs(
     
     if(do_bind_attrs)
     {
-        if(attr->resuseaddr)
+        if(attr->reuseaddr)
         {
             result = globus_xio_system_socket_setsockopt(
                 fd, SOL_SOCKET, SO_REUSEADDR, &int_one, sizeof(int_one));
@@ -1250,6 +1381,7 @@ globus_l_xio_tcp_create_listener(
     char                                portbuf[10];
     char *                              port;
     globus_xio_system_socket_t          fd;
+    globus_bool_t                       try_again = GLOBUS_FALSE;
     GlobusXIOName(globus_l_xio_tcp_create_listener);
     
     GlobusXIOTcpDebugEnter();
@@ -1267,110 +1399,156 @@ globus_l_xio_tcp_create_listener(
     /* setup hints for types of connectable sockets we want */
     memset(&addrinfo_hints, 0, sizeof(globus_addrinfo_t));
     addrinfo_hints.ai_flags = GLOBUS_AI_PASSIVE;
-    addrinfo_hints.ai_family = attr->no_ipv6 ? PF_INET : PF_UNSPEC;
     addrinfo_hints.ai_socktype = SOCK_STREAM;
     addrinfo_hints.ai_protocol = 0;
-    
-    result = globus_libc_getaddrinfo(
-        attr->bind_address, port, &addrinfo_hints, &save_addrinfo);
-    if(result != GLOBUS_SUCCESS && 
-        attr->listener_serv && attr->listener_port > 0)
+    if(attr->bind_address)
     {
-        /* it's possible the service name doesnt exist, since they also
-         * specified a numeric port, lets try that one
-         */
-        snprintf(portbuf, sizeof(portbuf), "%d", attr->listener_port);
-        result = globus_libc_getaddrinfo(
-            attr->bind_address, portbuf, &addrinfo_hints, &save_addrinfo);
+        addrinfo_hints.ai_family = attr->no_ipv6 ? PF_INET : PF_UNSPEC;
+    }
+    else
+    {
+        addrinfo_hints.ai_family = attr->no_ipv6 ? PF_INET : PF_INET6;
     }
 
-    if(result != GLOBUS_SUCCESS)
-    {
-        result = GlobusXIOErrorWrapFailed("globus_libc_getaddrinfo", result);
-        goto error_getaddrinfo;
-    }
-    
-    /* bind to the first one possible --
-     * its not possible to bind multiple interfaces (except on some openbsd)
-     * so, if the system doesnt map its inet interfaces to inet6, we
-     * may have a problem
-     */
-    for(addrinfo = save_addrinfo; addrinfo; addrinfo = addrinfo->ai_next)
-    {
-        if(GlobusLibcProtocolFamilyIsIP(addrinfo->ai_family))
+    /* getaddrinfo now returns results in the 'correct' rfc order of ipv4 
+     * first and ipv6 last, but we want to bind to ipv6 addresses first. 
+     * we'll loop through the whole getaddrinfo+bind twice -- first for V6, 
+     * then for V4 */
+    do
+    {   
+        addrinfo = NULL;
+         
+        result = globus_libc_getaddrinfo(
+            attr->bind_address, port, &addrinfo_hints, &save_addrinfo);
+        if(result != GLOBUS_SUCCESS && 
+            attr->listener_serv && attr->listener_port > 0)
         {
-            globus_bool_t               found = GLOBUS_FALSE;
-            
-            do
+            /* it's possible the service name doesnt exist, since they also
+             * specified a numeric port, lets try that one
+             */
+            snprintf(portbuf, sizeof(portbuf), "%d", attr->listener_port);
+            result = globus_libc_getaddrinfo(
+                attr->bind_address, portbuf, &addrinfo_hints, &save_addrinfo);
+        }
+
+        if(result != GLOBUS_SUCCESS)
+        {
+            if(addrinfo_hints.ai_family == PF_INET6)
             {
-                result = globus_xio_system_socket_create(
-                    &fd,
-                    addrinfo->ai_family,
-                    addrinfo->ai_socktype,
-                    addrinfo->ai_protocol);
-                if(result != GLOBUS_SUCCESS)
-                {
-                    break;
-                }
+                try_again = GLOBUS_TRUE;
+                addrinfo_hints.ai_family = PF_INET;
+                continue;
+            }
+            else
+            {
+                result = GlobusXIOErrorWrapFailed(
+                    "globus_libc_getaddrinfo", result);
+                goto error_getaddrinfo;
+            }
+        }
+        
+        for(addrinfo = save_addrinfo; addrinfo; addrinfo = addrinfo->ai_next)
+        {
+            if(GlobusLibcProtocolFamilyIsIP(addrinfo->ai_family))
+            {
+                globus_bool_t               found = GLOBUS_FALSE;
                 
-                result = globus_l_xio_tcp_apply_handle_attrs(
-                    attr, fd, GLOBUS_TRUE, GLOBUS_FALSE);
-                if(result != GLOBUS_SUCCESS)
+                do
                 {
-                    result = GlobusXIOErrorWrapFailed(
-                        "globus_l_xio_tcp_apply_handle_attrs", result);
-                    globus_xio_system_socket_close(fd);
-                    break;
-                }
-                
-                result = globus_l_xio_tcp_bind(
-                    fd,
-                    addrinfo->ai_addr,
-                    addrinfo->ai_addrlen,
-                    attr->restrict_port ? attr->listener_min_port : 0,
-                    attr->restrict_port ? attr->listener_max_port : 0,
-                    GLOBUS_TRUE);
-                if(result != GLOBUS_SUCCESS)
-                {
-                    result = GlobusXIOErrorWrapFailed(
-                        "globus_l_xio_tcp_bind", result);
-                    globus_xio_system_socket_close(fd);
-                    break;
-                }
-                
-                result = globus_xio_system_socket_listen(
-                    fd, 
-                    attr->listener_backlog < 0 
-                        ? SOMAXCONN : attr->listener_backlog);
-                if(result != GLOBUS_SUCCESS)
-                {
-                    globus_xio_system_socket_close(fd);
-                    
-                    if(globus_error_errno_match(
-                        globus_error_peek(result),
-                        GLOBUS_XIO_MODULE,
-                        EADDRINUSE))
-                    {
-                        /* there's a race between bind and listen, let's try
-                         * this all over again
-                         */
-                        continue;
-                    }
-                    else
+                    result = globus_xio_system_socket_create(
+                        &fd,
+                        addrinfo->ai_family,
+                        addrinfo->ai_socktype,
+                        addrinfo->ai_protocol);
+                    if(result != GLOBUS_SUCCESS)
                     {
                         break;
                     }
-                }
+                    
+                    if(addrinfo->ai_family == AF_INET6 && !attr->bind_address)
+                    {
+                        int             int_zero = 0;
+                        
+                        result = globus_xio_system_socket_setsockopt(
+                            fd, IPPROTO_IPV6, IPV6_V6ONLY, &int_zero, sizeof(int_zero));
+                        if(result != GLOBUS_SUCCESS)
+                        {
+                             GlobusXIOTcpDebugPrintf(
+                                GLOBUS_L_XIO_TCP_DEBUG_INFO,
+                                ("Unable to set V6ONLY sockopt."));
+                        }
+                    }
+
+                    result = globus_l_xio_tcp_apply_handle_attrs(
+                        attr, fd, GLOBUS_TRUE, GLOBUS_FALSE);
+                    if(result != GLOBUS_SUCCESS)
+                    {
+                        result = GlobusXIOErrorWrapFailed(
+                            "globus_l_xio_tcp_apply_handle_attrs", result);
+                        globus_xio_system_socket_close(fd);
+                        break;
+                    }
+                    
+                    result = globus_l_xio_tcp_bind(
+                        fd,
+                        addrinfo->ai_addr,
+                        addrinfo->ai_addrlen,
+                        attr->restrict_port ? attr->listener_min_port : 0,
+                        attr->restrict_port ? attr->listener_max_port : 0,
+                        GLOBUS_TRUE);
+                    if(result != GLOBUS_SUCCESS)
+                    {
+                        result = GlobusXIOErrorWrapFailed(
+                            "globus_l_xio_tcp_bind", result);
+                        globus_xio_system_socket_close(fd);
+                        break;
+                    }
+                    
+                    result = globus_xio_system_socket_listen(
+                        fd, 
+                        attr->listener_backlog < 0 
+                            ? SOMAXCONN : attr->listener_backlog);
+                    if(result != GLOBUS_SUCCESS)
+                    {
+                        globus_xio_system_socket_close(fd);
+                        
+                        if(globus_error_errno_match(
+                            globus_error_peek(result),
+                            GLOBUS_XIO_MODULE,
+                            EADDRINUSE))
+                        {
+                            /* there's a race between bind and listen, let's try
+                             * this all over again
+                             */
+                            continue;
+                        }
+                        else
+                        {
+                            break;
+                        }
+                    }
+                    
+                    found = GLOBUS_TRUE;
+                } while(!found);
                 
-                found = GLOBUS_TRUE;
-            } while(!found);
-            
-            if(found)
-            {
-                break;
+                if(found)
+                {
+                    break;
+                }
             }
         }
-    }
+        
+        if(!addrinfo && addrinfo_hints.ai_family == PF_INET6)
+        {
+            try_again = GLOBUS_TRUE;
+            addrinfo_hints.ai_family = PF_INET;
+            globus_libc_freeaddrinfo(save_addrinfo);
+        }
+        else
+        {
+            try_again = GLOBUS_FALSE;
+        }
+    } while(!addrinfo && try_again);
     
     if(!addrinfo)
     {
@@ -2725,6 +2903,102 @@ globus_l_xio_tcp_cntl(
         *out_bool = handle->use_blocking_io;
         break;
         
+      case GLOBUS_XIO_GET_STRING_OPTIONS:
+      {
+        size_t string_opts_len = 1;
+        globus_bool_t keepalive, oobinline, nodelay;
+        int sndbuf, rcvbuf;
+        out_string = va_arg(ap, char **);
+
+        *out_string = NULL;
+
+        len = sizeof(globus_bool_t);
+        result = globus_xio_system_socket_getsockopt(
+            fd, SOL_SOCKET, SO_KEEPALIVE, &keepalive, &len);
+        if (result)
+        {
+            goto error_sockopt;
+        }
+
+        len = sizeof(globus_bool_t);
+        result = globus_xio_system_socket_getsockopt(
+            fd, SOL_SOCKET, SO_OOBINLINE, &oobinline, &len);
+        if (result)
+        {
+            goto error_sockopt;
+        }
+
+        len = sizeof(int);
+        result = globus_xio_system_socket_getsockopt(
+            fd, SOL_SOCKET, SO_SNDBUF, &sndbuf, &len);
+        if (result)
+        {
+            goto error_sockopt;
+        }
+
+        len = sizeof(int);
+        result = globus_xio_system_socket_getsockopt(
+            fd, SOL_SOCKET, SO_RCVBUF, &rcvbuf, &len);
+        if (result)
+        {
+            goto error_sockopt;
+        }
+
+        len = sizeof(globus_bool_t);
+        result = globus_xio_system_socket_getsockopt(
+            fd, IPPROTO_TCP, TCP_NODELAY, &nodelay, &len);
+        if (result)
+        {
+            goto error_sockopt;
+        }
+
+        string_opts_len += snprintf(
+                NULL, 0,
+                "keepalive=%s;",
+                keepalive ? "true" : "false");
+        string_opts_len += snprintf(
+                NULL, 0,
+                "oobinline=%s;",
+                oobinline ? "true" : "false");
+        string_opts_len += snprintf(
+                NULL, 0,
+                "sndbuf=%d;",
+                sndbuf);
+        string_opts_len += snprintf(
+                NULL, 0,
+                "rcvbuf=%d;",
+                rcvbuf);
+        string_opts_len += snprintf(
+                NULL, 0,
+                "nodelay=%s;",
+                nodelay ? "true" : "false");
+        *out_string = malloc(string_opts_len);
+
+        string_opts_len = 0;
+        string_opts_len += sprintf(
+                *out_string + string_opts_len,
+                "keepalive=%s;",
+                keepalive ? "true" : "false");
+        string_opts_len += sprintf(
+                *out_string + string_opts_len,
+                "oobinline=%s;",
+                oobinline ? "true" : "false");
+        string_opts_len += sprintf(
+                *out_string + string_opts_len,
+                "sndbuf=%d;",
+                sndbuf);
+        string_opts_len += sprintf(
+                *out_string + string_opts_len,
+                "rcvbuf=%d;",
+                rcvbuf);
+        string_opts_len += sprintf(
+                *out_string + string_opts_len,
+                "nodelay=%s;",
+                nodelay ? "true" : "false");
+        *((*out_string) + string_opts_len - 1) = '\0';
+      }
+      break;
+        
       default:
         result = GlobusXIOErrorInvalidCommand(cmd);
         goto error_invalid;
@@ -2788,6 +3062,10 @@ globus_l_xio_tcp_init(
     globus_xio_driver_string_cntl_set_table(
         driver,
         tcp_l_string_opts_table);
+
+    globus_xio_driver_handle_string_cntl_set_table(
+        driver,
+        tcp_l_handle_string_opts_table);
     
     *out_driver = driver;
     
