@@ -15,7 +15,7 @@
  */
 
 
-#include "config.h"
+#include "globus_i_xio_config.h"
 #include "globus_common.h"
 #ifndef TARGET_ARCH_WIN32
 #include "globus_xio_system.h"
@@ -568,7 +568,7 @@ globus_l_xio_system_cancel_cb(
                         {
                             globus_list_insert(
                                 &globus_l_xio_system_canceled_reads,
-                                (void *) op_info->handle->fd);
+                                (void *) (intptr_t) op_info->handle->fd);
                         }
                         else
                         {
@@ -582,7 +582,7 @@ globus_l_xio_system_cancel_cb(
                         {
                             globus_list_insert(
                                 &globus_l_xio_system_canceled_writes,
-                                (void *) op_info->handle->fd);
+                                (void *) (intptr_t) op_info->handle->fd);
                         }
                         else
                         {
@@ -1217,7 +1217,7 @@ error_canceled:
     return handled_it;
 }
 
-/**
+/*
  * one of these fds is bad, lock down the fdset and check them all
  * --- assumed to be called with cancel lock held after a select
  */
@@ -1254,7 +1254,8 @@ globus_l_xio_system_bad_apple(void)
                         op_info->state = GLOBUS_I_XIO_SYSTEM_OP_CANCELED;
                         op_info->error = GlobusXIOErrorObjParameter("handle");
                         globus_list_insert(
-                            &globus_l_xio_system_canceled_reads, (void *) fd);
+                            &globus_l_xio_system_canceled_reads,
+                            (void *) (intptr_t) fd);
                     }
                 }
             }
@@ -1276,7 +1277,8 @@ globus_l_xio_system_bad_apple(void)
                         op_info->state = GLOBUS_I_XIO_SYSTEM_OP_CANCELED;
                         op_info->error = GlobusXIOErrorObjParameter("handle");
                         globus_list_insert(
-                            &globus_l_xio_system_canceled_writes, (void *) fd);
+                            &globus_l_xio_system_canceled_writes,
+                            (void *) (intptr_t) fd);
                     }
                 }
             }
@@ -1386,7 +1388,7 @@ globus_l_xio_system_poll(
                     globus_l_xio_system_bad_apple();
                 }
                 
-                /**
+                /*
                  * can't really do anything about other errors
                  * so, set ready fds to known state in case there are things
                  * to be canceled
@@ -1404,7 +1406,7 @@ globus_l_xio_system_poll(
 
             while(!globus_list_empty(globus_l_xio_system_canceled_reads))
             {
-                fd = (int) globus_list_remove(
+                fd = (int) (intptr_t) globus_list_remove(
                     &globus_l_xio_system_canceled_reads,
                     globus_l_xio_system_canceled_reads);
                 
@@ -1421,7 +1423,7 @@ globus_l_xio_system_poll(
 
             while(!globus_list_empty(globus_l_xio_system_canceled_writes))
             {
-                fd = (int) globus_list_remove(
+                fd = (int) (intptr_t) globus_list_remove(
                     &globus_l_xio_system_canceled_writes,
                     globus_l_xio_system_canceled_writes);
                 
@@ -1931,7 +1933,7 @@ globus_l_xio_system_read(
         globus_size_t                   nbytes = *u_nbytes;
         globus_size_t                   total = nbytes;
         
-        /**
+        /*
          * XXX this is not thread safe... both reads and writes are mucking
          * with blocking status.
          * worst case, we read 0 bytes in the loop below, return, and xio
@@ -2081,7 +2083,7 @@ globus_l_xio_system_write(
         globus_size_t                   nbytes = *u_nbytes;
         globus_size_t                   total = nbytes;
         
-        /**
+        /*
          * XXX this is not thread safe... both reads and writes are mucking
          * with blocking status
          */
